#!/usr/bin/env python
from __future__ import print_function
from __future__ import unicode_literals
from collections import OrderedDict
import re
import consul
import requests


class ConsulManagerException(Exception):
    """Applications can catch all library exceptions using this one"""
    pass


class ConsulManagerInvalidRegex(ConsulManagerException):
    pass


class ConsulManagerInvalidEnv(ConsulManagerException):
    pass

class ConsulManagerInvalidKVStore(ConsulManagerException):
    pass

class ConsulManagerInvalidFormat(ConsulManagerException):
    """Consul returned a different format from the one we expect"""
    pass


class ConsulManagerInvalidParams(ConsulManagerException):
    pass


class ConsulManagerConnectionError(ConsulManagerException):
    pass


class ConsulManagerInvalidCloudSetup(ConsulManagerException):
    pass


class ConsulManagerIncompleteOperation(ConsulManagerException):
    pass


class ConsulManager(object):

    def __init__(self, host='localhost', port=8500, consistency='default'):

        self.consul = consul.Consul(host=host, port=port, consistency=consistency)
        self.environments_per_cloud = {}
        try:
            self.data_centers = self.consul.catalog.datacenters()
        except requests.exceptions.ConnectionError:
            raise ConsulManagerConnectionError("consul API is not reachable on {}:{}".format(host, port))

    def list_services(self, cloud=None):
        """ list services in one, several or all clouds
        Args:
            cloud (unicode or list): None for all clouds
                                     accepts a list if strings or single cloud_name
        Returns:
            dict: maintains python-consul dict format for services
                dict [service_name]:[<tags>]
        Raises:
            ConsulManagerConnectionError
        """
        clouds = self._get_clouds(cloud)
        all_services = {}
        for dc in clouds:
            _, services = self.consul.catalog.services(dc=dc)
            # merge two dicts of type <str>:[list]
            for serv in services:
                if serv in all_services:
                    all_services[serv] = list(set(all_services[serv] + services[serv]))
                else:
                    all_services[serv] = services[serv]

        return all_services

    def list_nodes(self, cloud=None, name=None, name_strict_match=None, service=None, service_tag=None, passing=None, environment=None):
        """ list nodes filtered by the provided criteria with AND operator between them
        Args:
            cloud       (unicode or list):  None for all clouds
                                        accepts a list if strings or single cloud_name
            name        (unicode): used to filter by the node's FQDN, can be any valid python regex
            name_strict_match (bool):   if true, name will not be matched as regex
            service     (unicode or list): service name, required if `passing` is used
            service_tag (unicode): service_tag name
            passing     (bool)   : True returns only healthy nodes, False returns non healthy ones
                                   passing parameter is tied to a provided service parameter
            environment (unicode): one of production, staging, testing, development

        Returns:
            OrderedDict: node[<Node_name>]: {address:<ip_address>,tags:<[]>}
        Raises:
            ConsulManagerInvalidRegex
            ConsulManagerInvalidEnv
            ConsulManagerInvalidFormat
            ConsulManagerInvalidParams
            ConsulManagerConnectionError
            ConsulManagerInvalidCloudSetup
        """
        try:
            if passing and not service:
                raise ConsulManagerInvalidParams("`service` parameter is required when `passing` is provided")

            all_nodes = []
            clouds = self._get_clouds(cloud)
            if environment:
                node_meta = {'environment': environment}
            else:
                node_meta = None

            if service or service_tag:
                all_nodes = self._nodes_per_service_or_service_tag(clouds, service, service_tag, passing, node_meta)
            else:
                # list all nodes
                for dc in clouds:
                    if environment:
                        _, nodes = self.consul.catalog.nodes(dc=dc, node_meta=node_meta)
                    else:
                        _, nodes = self.consul.catalog.nodes(dc=dc)

                    all_nodes.extend(nodes)

            all_nodes = self._normalize_nodes(all_nodes)

            # filter by `name`
            result = OrderedDict()
            if name:
                if name_strict_match:
                    for node_name, node_attributes in all_nodes.items():
                        if name == node_name:
                            result[node_name] = node_attributes

                else:  # filter by `name` as a regex
                    try:
                        reg = re.compile(name)
                    except Exception as ex:
                        raise ConsulManagerInvalidRegex("Invalid regular expression: {}".format(ex))

                    for node_name, node_attributes in all_nodes.items():
                        if reg.match(node_name):
                            result[node_name] = node_attributes
            else:
                result = all_nodes

            # sort by FQDN
            try:
                ret = OrderedDict(sorted(result.items(), key=lambda fqdn: fqdn[0].split('.')[1]))
            except IndexError as ex:
                ret = OrderedDict(sorted(result.items()))
        except requests.exceptions.ConnectionError as ex:
            raise ConsulManagerConnectionError("consul API is not reachable: {}".format(ex))

        return ret

    def _get_clouds(self, cloud):
        """ helper function to interpret cloud parameter as all (default), a list or only one cloud"""
        clouds = []
        if not cloud:
            clouds = self.data_centers
        elif isinstance(cloud, list):
            clouds.extend(cloud)
        else:
            clouds = [cloud]
        return clouds

    def _nodes_per_service_or_service_tag(self, clouds, service, service_tag, passing, node_meta):
        """service could be a list of service names or a single unicode service  name argument
        when a list is provided an AND relation is assumed
        """
        ret = []
        for dc in clouds:
            all_nodes = []
            if service:  # service was specified => service_tag will specify a subset of service result
                if isinstance(service, list):  # do AND between the results of several queries
                    nodes = self._and_nodes_from_multiple_services(service, dc, passing, node_meta)
                else:
                    _, nodes = self.consul.health.service(service, dc=dc, passing=passing, node_meta=node_meta)
                """
                consul api only supports passing = None or True (False is equivalent None in consul api)
                so to get non healthy nodes we should subtract healthy from all
                """
                if passing == False:  # must equal False, we cant use `is` here. ignore E712
                    _, healthy_nodes = self.consul.health.service(service, dc=dc, passing=True, node_meta=node_meta)
                    nodes = [n for n in nodes if n not in healthy_nodes]
                if service_tag:
                    nodes = self._filter_nodes_by_service_tag(nodes, service_tag)
                all_nodes.extend(nodes)

            else:  # service was not specified => we should look up `service_tag` in all services
                _, services = self.consul.catalog.services(dc=dc, node_meta=node_meta)
                for serv in services.keys():
                    _, nodes = self.consul.health.service(serv, dc=dc, passing=passing, node_meta=node_meta)
                    if passing == False:
                        _, healthy_nodes = self.consul.health.service(serv, dc=dc, passing=True, node_meta=node_meta)
                        nodes = [n for n in nodes if n not in healthy_nodes]
                    nodes = self._filter_nodes_by_service_tag(nodes, service_tag)
                    all_nodes.extend(nodes)

            ret.extend(all_nodes)

        return ret

    def _and_nodes_from_multiple_services(self, service, dc, passing, node_meta):
        """return nodes element that exist in *all* the services"""
        nodes_all_services = []
        nodes_dict = {}
        nodes = []
        # get a list of nodes of all provided services
        for srv in service:
            _, nodes_by_service = self.consul.health.service(srv, dc=dc, passing=passing, node_meta=node_meta)
            nodes_all_services.extend(nodes_by_service)

        # create a dictionary with node and count
        for n in nodes_all_services:
            if nodes_dict.get(n.get('Node').get('Node')):  # no exception handling, consul endpoint should not change the format
                nodes_dict[n.get('Node').get('Node')] = (n, nodes_dict[n.get('Node').get('Node')][1] + 1)
            else:
                nodes_dict[n.get('Node').get('Node')] = (n, 1)

        # get the nodes that exist in all nodes_for_service
        for node_count in nodes_dict.values():
            if node_count[1] == len(service):
                nodes.append(node_count[0])
        return nodes

    def _get_dc_of_node(self, node_name):
        for dc in self.data_centers:
            if self.list_nodes(cloud=dc, name=node_name, name_strict_match=True):
                return dc

    def add_node_to_kv(self, node_name, store_name):
        """ Add the provided node to consul's shared kv store of the corresponding cloud
        Raises:
            ConsulManagerInvalidKVStore
            ConsulManagerException
        """
        dc_of_node = self._get_dc_of_node(node_name)
        if not dc_of_node:
            raise ConsulManagerException("Couldn't locate {} in any cloud".format(node_name))
        self._modify_consul_shared_kv('add', [node_name], store_name, dc=dc_of_node)

    def remove_node_from_kv(self, node_name, store_name):
        """ Remove the provided node from consul's shared kv store of the corresponding cloud
        Raises:
            ConsulManagerInvalidKVStore
        """
        dc_of_node = self._get_dc_of_node(node_name)
        self._modify_consul_shared_kv('remove', [node_name], store_name, dc=dc_of_node)

    def _modify_consul_shared_kv(self, action, node_list, store_name, dc=None):
        """ Add or remove a list from/to consul's shared kv store
        Args:
            action:     (unicode):  add/remove
            nodes:      (list):     list of nodes names to modify
            store_name  (unicode):  name of kv store at consul
            dc          (unicode):  cloud name where kv store is located
        Returns:
            number of modified nodes
        """
        _, response = self.consul.kv.get(store_name, dc=dc)
        if not response:
            raise ConsulManagerInvalidKVStore("Unknown KV store {}".format(store_name))
        if response.get('Value'):
            current_items = response.get('Value').decode().replace("'", "").split(',')
        else:
            current_items = []
        size_before = len(current_items)
        items_to_modify = 0
        if action == 'remove':
            for node in node_list:
                if node in current_items:
                    current_items.remove(node)
                    items_to_modify += 1
        else:  # action == 'add'
            for node in node_list:
                if node not in current_items:
                    current_items.append(node)
                    items_to_modify += 1
        self.consul.kv.put(store_name, str(",".join(current_items)), dc=dc)

        # check the operation went correctly
        _, response = self.consul.kv.get(store_name, dc=dc)
        if response.get('Value'):
            items_after = response.get('Value').decode().replace("'", "").split(',')
        else:
            items_after = []
        size_after = len(items_after)
        if abs(size_after - size_before) != items_to_modify:
            raise ConsulManagerIncompleteOperation("Inconsistent kv modify operation?")
        return items_to_modify

    @staticmethod
    def _detect_format(result):
        """ helper to detect returned format from consul
        health.service returns a list of two level depth dict{'Checks':{}, 'Node':{}, 'Service':{}}
        catalog.service/nodes returns a list of one level depth dict{'Node':'',...}"""
        if len(result) > 0 and 'Checks' in result[0]:
            return 'health'
        else:
            return 'service'

    @staticmethod
    def _normalize_nodes(result):
        """ normalizes data returned by different api calls by creating unique
        data structure that selects interesting data from each call

        Args:
            result (list): a list of dicts, format dicts differs depending on used api call
        Returns:
            OrderedDict: node[<Node_name>]: {address:<ip_address>,tags:<[]>}
        """
        ret = OrderedDict()
        if ConsulManager._detect_format(result) == 'health':
            for item in result:
                node = item.get('Node')
                if node:
                    ret[node.get('Node')] = {'address': node.get('Address'),
                                             'tags': []}  # TODO: pending Consul upgrade
        else:
            for item in result:
                node = item.get('Node')
                if node:
                    ret[node] = {'address': item.get('Address'),
                                 'tags': []}  # TODO: pending Consul upgrade
        return ret

    @staticmethod
    def _filter_nodes_by_service_tag(nodes, service_tag):
        filtered_by_tag = []
        for node in nodes:
            try:
                if service_tag in node.get('Service').get('Tags'):
                    filtered_by_tag.append(node)
            except TypeError as ex:
                raise ConsulManagerInvalidFormat("health.service should return dict[Service][Tags]: {}".format(ex))

        return filtered_by_tag


if __name__ == "__main__":
    import pprint
    c = ConsulManager()

    print("list_services()")
    n = c.list_services()
    pprint.pprint(n)

    service_name = 'consul'
    node_name = 'api'

    print("list_nodes(service='{}')".format(service_name))
    n = c.list_nodes(service=service_name)
    pprint.pprint(n)

    print("list_nodes(service='{}',passing='{}')".format(service_name, True))
    n = c.list_nodes(service=service_name, passing=True)
    pprint.pprint(n)

    print("list_nodes(service='{}',passing='{}')".format(service_name, False))
    n = c.list_nodes(service=service_name, passing=False)
    pprint.pprint(n)

    print("list_nodes(service='invalid-service-name')")
    n = c.list_nodes(service='invalid-service-name')
    pprint.pprint(n)

    print("list_nodes(name='{}')".format(node_name))
    n = c.list_nodes(name=node_name)
    pprint.pprint(n)

    print("list_nodes(name='{}', service='{}')".format(node_name + '01', service_name))
    n = c.list_nodes(name=node_name + '01', service=service_name)
    pprint.pprint(n)

    print("list_nodes(service_tag='traffic=default')")
    n = c.list_nodes(service_tag='traffic=default')
    pprint.pprint(n)

    print("list_nodes(service='ui-server', service_tag='traffic=default')")
    n = c.list_nodes(service='ui-server', service_tag='traffic=default')
    pprint.pprint(n)
